from cmath import pi
from collections import OrderedDict
from typing import Tuple, List, Optional
from pathlib import Path
from random import randint, shuffle
import numpy as np
from sklearn.kernel_ridge import KernelRidge
from scipy import stats
from dovado_rtl.enums import RegressionModel
from sklearn.model_selection import LeaveOneOut, RandomizedSearchCV
from dovado_rtl.config import Configuration
from dovado_rtl.abstract_classes import (
    AbstractEstimator,
    AbstractDesignPointEvaluator,
)
from dovado_rtl.simple_types import Example, Metric


class Estimator(AbstractEstimator):
    def __init__(
        self,
        regression_model: RegressionModel,
        design_point_evaluator: AbstractDesignPointEvaluator,
        free_parameters_range: "OrderedDict[str, Tuple[int, int]]",
        dataset_size: int,
        config: Configuration,
    ):
        self.__regression_model = regression_model
        self.__examples: List[Example] = []
        self.__examples_updated: bool = True
        self.__estimator = None
        self.__design_point_evaluator: AbstractDesignPointEvaluator = design_point_evaluator
        self.__dataset_size: int = dataset_size
        self.__config: Configuration = config
        self.__design_point_evaluator.set_estimator(self)
        self.__metrics: Optional[List[Metric]] = None
        if self.__config.get_config("ESTIMATION_TESTING"):
            self.__fname = str(self.__config.get_config("WORK_DIR")) + str(
                self.__config.get_config("EST_TEST_CSV")
            )
            Path(self.__fname).open("w").close()
        self.__generate_dataset(free_parameters_range,)

    def estimate(self, design_point: List[int], metric: Metric) -> float:
        if self.__regression_model is RegressionModel.KERNEL_RIDGE:
            estimate = self.__kernel_ridge(design_point, metric)
        else:
            raise Exception("Other models still need to be implemented")

        if self.__config.get_config("ESTIMATION_TESTING") and len(
            self.__examples
        ) > max(len(list(design_point)), 5):
            evaluated = self.__design_point_evaluator.evaluate(
                tuple(design_point)
            )
            line = (
                str(evaluated.value[metric])
                + ","
                + str(estimate)
                + ","
                + (
                    str(metric.utilisation[0])
                    + "-"
                    + str(metric.utilisation[1])
                    if not metric.is_frequency
                    else "Frequency"
                )
            )

            Path(self.__fname).open("a").writelines([line + "\n"])

        return estimate

    def __kernel_ridge(self, design_point: List[int], metric: Metric) -> float:
        X = self.__get_independent_variables()
        X = X.astype("int64").reshape(-1, len(design_point))
        y = self.__get_dependent_variables(metric)
        y = y.astype("float64").reshape(-1, 1)
        reg = KernelRidge()
        parameters = {
            "alpha": stats.reciprocal(a=1e-7, b=1e2, size=10),
            "gamma": stats.expon(scale=0.1),
            "kernel": ["rbf"],
        }
        cv = LeaveOneOut()
        cv_reg = RandomizedSearchCV(
            reg,
            parameters,
            cv=cv,
            scoring="neg_root_mean_squared_error",
            n_jobs=-1,
        )
        cv_reg.fit(X, y)
        design_point = np.reshape(design_point, (1, -1))
        return cv_reg.predict(design_point)[0][0]

    def add_example(self, example: Example) -> None:
        if not self.__metrics:
            self.__metrics = self.__design_point_evaluator.get_metrics()
        self.__examples.append(example)
        shuffle(self.__examples)
        self.__examples_updated = True

    def __generate_dataset(
        self, parameters_range: "OrderedDict[str, Tuple[int, int]]",
    ) -> None:
        for _ in range(0, self.__dataset_size):
            design_point = []
            for k in parameters_range.keys():
                design_point.append(
                    randint(parameters_range[k][0], parameters_range[k][1])
                )
            design_value = self.__design_point_evaluator.evaluate(
                tuple(design_point)
            )
            self.add_example(
                Example(design_point=design_point, design_value=design_value)
            )

    def __get_dependent_variables(self, metric: Metric) -> "np.ndarray":
        dependent_variable = []
        for example in self.__examples:
            dependent_variable.append(example.design_value.value[metric])
        return np.array(dependent_variable)

    def __get_independent_variables(self) -> "np.ndarray":
        independent_variables = []
        for example in self.__examples:
            independent_variables.append(example.design_point)
        return np.array(independent_variables)

    def get_examples(self):
        return self.__examples



from pathlib import Path
import dovado_rtl.user_input as user_input
from dovado_rtl.config import Configuration
import dovado_rtl.vivado_interaction as vivado
from dovado_rtl.frame_handling import (
    TclFrameHandler,
    HdlBoxFrameHandler,
    XdcFrameHandler,
)
from dovado_rtl.enums import RTL, RegressionModel
from dovado_rtl.point_evaluation import DesignPointEvaluator
from dovado_rtl.estimation import Estimator
from dovado_rtl.fitness import FitnessEvaluator
from dovado_rtl.genetic_algorithm import optimize


def main():
    vivado.start()
    config = Configuration()
    Path(str(config.get_config("WORK_DIR"))).mkdir(parents=True, exist_ok=True)

    src_folder = user_input.ask_code_dir()
    top_module, _, parsed_source = user_input.ask_top_module(
        src_folder, config
    )
    parsed_source.set_entity(top_module.get_name())
    synthesis_part = user_input.ask_part()

    stop_step = user_input.ask_stop_step()

    free_parameters = user_input.ask_parameters(parsed_source)

    clock_port = user_input.ask_identifiers(parsed_source)

    incremental_mode = user_input.ask_incremental_mode(stop_step)

    (
        synthesis_directive,
        place_directive,
        route_directive,
    ) = user_input.ask_directives(stop_step, incremental_mode)

    # Clock given in Mhz
    target_clock = user_input.ask_clock()

    XdcFrameHandler(
        str(config.get_config("PLACEHOLDER")),
        str(config.get_config("XDC_DIR"))
        + str(config.get_config("CONSTRAINT_FRAME")),
        1000 / target_clock,
        str(config.get_config("WORK_DIR"))
        + str(config.get_config("CONSTRAINT")),
    ).fill()

    tcl_handler = TclFrameHandler(
        config,
        parsed_source,
        src_folder,
        synthesis_part,
        synthesis_directive,
        incremental_mode,
        stop_step,
        place_directive,
        route_directive,
    )
    tcl_handler.fill()

    box_handler = HdlBoxFrameHandler(
        str(config.get_config("PLACEHOLDER")),
        str(config.get_config("VHDL_DIR"))
        + str(config.get_config("VHDL_BOX_FRAME"))
        if parsed_source.get_hdl() is RTL.VHDL
        else str(config.get_config("VERILOG_DIR"))
        + str(config.get_config("VERILOG_BOX_FRAME")),
        top_module.get_name(),
        parsed_source.get_ports(),
        parsed_source.get_port(clock_port.get_name()),
        str(config.get_config("WORK_DIR")) + str(config.get_config("VHDL_BOX"))
        if parsed_source.get_hdl() is RTL.VHDL
        else str(config.get_config("WORK_DIR"))
        + str(config.get_config("VERILOG_BOX")),
        parsed_source.get_hdl(),
        parsed_source.get_folder()
        if parsed_source.get_hdl() is RTL.VHDL
        else None,
    )

    point_evaluator = DesignPointEvaluator(
        config,
        parsed_source,
        box_handler,
        tcl_handler,
        target_clock,
        incremental_mode,
        stop_step,
        [p.get_name() for p in free_parameters],
    )

    if user_input.ask_is_point_evaluation():
        print(
            "Point Evaluation result\n"
            + str(
                point_evaluator.evaluate(
                    tuple(user_input.ask_parameters_value(free_parameters))
                )
            )
        )
        return

    free_parameters_range = user_input.ask_parameters_range(
        free_parameters, config
    )

    estimator = Estimator(
        RegressionModel.KERNEL_RIDGE,
        point_evaluator,
        free_parameters_range,
        int(config.get_config("INITIAL_SAMPLES")),
        config,
    )

    fitness_evaluator = FitnessEvaluator(estimator, point_evaluator, config)

    metrics = point_evaluator.get_metrics()

    if not metrics:
        raise Exception(
            "INITIAL_SAMPLES must be at least one in order to analyze "
            + "which utilization metrics are available for the given board"
        )

    result = optimize(
        fitness_evaluator,
        free_parameters_range,
        metrics,
        str(config.get_config("GENETIC_RUN_TIME")),
    )
    print("Optimization Result: " + str(result))


main()
